/*This file has been prepared for Doxygen automatic documentation generation.*/
/*! \file *********************************************************************
 *
 * \brief FreeRTOS Real Time Kernel example.
 *
 * Creates all the demo application tasks, then starts the scheduler.  The WEB
 * documentation provides more details of the demo application tasks.
 *
 * Main. c also creates a task called "Check".  This only executes every three
 * seconds but has the highest priority so is guaranteed to get processor time.
 * Its main function is to check that all the other tasks are still operational.
 * Each task that does not flash an LED maintains a unique count that is
 * incremented each time the task successfully completes its function.  Should
 * any error occur within such a task the count is permanently halted.  The
 * check task inspects the count of each task to ensure it has changed since
 * the last time the check task executed.  If all the count variables have
 * changed all the tasks are still executing error free, and the check task
 * toggles an LED.  Should any task contain an error at any time the LED toggle
 * will stop.
 *
 * The LED flash and communications test tasks do not maintain a count.
 *
 * - Compiler:           IAR EWAVR32 and GNU GCC for AVR32
 * - Supported devices:  All AVR32 devices with GPIO.
 * - AppNote:
 *
 * \author               Atmel Corporation: http://www.atmel.com \n
 *                       Support and FAQ: http://support.atmel.no/
 *
 *****************************************************************************/

/*
    FreeRTOS V7.0.0 - Copyright (C) 2011 Real Time Engineers Ltd.
	

    ***************************************************************************
     *                                                                       *
     *    FreeRTOS tutorial books are available in pdf and paperback.        *
     *    Complete, revised, and edited pdf reference manuals are also       *
     *    available.                                                         *
     *                                                                       *
     *    Purchasing FreeRTOS documentation will not only help you, by       *
     *    ensuring you get running as quickly as possible and with an        *
     *    in-depth knowledge of how to use FreeRTOS, it will also help       *
     *    the FreeRTOS project to continue with its mission of providing     *
     *    professional grade, cross platform, de facto standard solutions    *
     *    for microcontrollers - completely free of charge!                  *
     *                                                                       *
     *    >>> See http://www.FreeRTOS.org/Documentation for details. <<<     *
     *                                                                       *
     *    Thank you for using FreeRTOS, and thank you for your support!      *
     *                                                                       *
    ***************************************************************************


    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.
    >>>NOTE<<< The modification to the GPL is included to allow you to
    distribute a combined work that includes FreeRTOS without being obliged to
    provide the source code for proprietary components outside of the FreeRTOS
    kernel.  FreeRTOS is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details. You should have received a copy of the GNU General Public
    License and the FreeRTOS license exception along with FreeRTOS; if not it
    can be viewed here: http://www.freertos.org/a00114.html and also obtained
    by writing to Richard Barry, contact details for whom are available on the
    FreeRTOS WEB site.

    1 tab == 4 spaces!

    http://www.FreeRTOS.org - Documentation, latest information, license and
    contact details.

    http://www.SafeRTOS.com - A version that is certified for use in safety
    critical systems.

    http://www.OpenRTOS.com - Commercial support, development, porting,
    licensing and training services.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <intc.h>
#include <interrupt.h>
#include <sleepmgr.h>
#include <uc3/sleepmgr.h>
#include <gpio.h>
#include <flashc.h>
#include <power_clocks_lib.h>
#include <sleep.h>
#include <sysclk.h>
#include <tc.h>
#include <usart.h>
#include <udi_cdc.h>
#include <usb_protocol_cdc.h>
#include <udc.h>
#include <udd.h>
#include <compiler.h>
#include <status_codes.h>
#include <board.h>
#include <stdio_usb.h>


// defines for USB UART
#define CONFIG_USART_IF (AVR32_USART2)

// include FreeRTOS header files
#include "FreeRTOS.h"
#include "task.h"

// define board in AVR32/GNU C Compiler - Symbols - Defined Symbols (-D)
// BOARD=UC3_A3_XPLAINED
// BOARD=EVK1104

#if BOARD == EVK1104
	#define BOARD_IS_EVK1104
#elif BOARD == UC3_A3_XPLAINED
	#define BOARD_IS_UC3_A3_XPLAINED
#endif


void init(void)
{
	// board init
	board_init();
	
#if UC3L
	// clock frequencies
	#define EXAMPLE_TARGET_DFLL_FREQ_HZ     96000000  // DFLL target frequency, in Hz
	#define EXAMPLE_TARGET_MCUCLK_FREQ_HZ   12000000  // MCU clock target frequency, in Hz
	#define EXAMPLE_TARGET_PBACLK_FREQ_HZ   12000000  // PBA clock target frequency, in Hz

	// parameters to pcl_configure_clocks().
	static scif_gclk_opt_t gc_dfllif_ref_opt = { SCIF_GCCTRL_SLOWCLOCK, 0, false};
	static pcl_freq_param_t pcl_dfll_freq_param = {
		.main_clk_src = PCL_MC_DFLL0,
		.cpu_f        = EXAMPLE_TARGET_MCUCLK_FREQ_HZ,
		.pba_f        = EXAMPLE_TARGET_PBACLK_FREQ_HZ,
		.pbb_f        = EXAMPLE_TARGET_PBACLK_FREQ_HZ,
		.dfll_f       = EXAMPLE_TARGET_DFLL_FREQ_HZ,
		.pextra_params = &gc_dfllif_ref_opt
	};
	pcl_configure_clocks(&pcl_dfll_freq_param);
#else
	pcl_switch_to_osc(PCL_OSC0, FOSC0, OSC0_STARTUP);
#endif	
	
	// stdio init
	stdio_usb_init(&CONFIG_USART_IF);

	// Specify that stdout and stdin should not be buffered.

#if defined(__GNUC__) && defined(__AVR32__)
	setbuf(stdout, NULL);
	setbuf(stdin,  NULL);
#endif
}

#include "edac.h"

#include <queue.h>
#include <semphr.h>

#define queueLength 4
struct AMessage
{
	portCHAR ucMessageID;
	uint8_t qId;	// only use last 5 bit
	uint8_t qAddr;
	uint8_t qData[5];
} xMessage;

struct ARequest
{
	uint8_t qId;	// only use last 5 bit
	uint8_t qAddr;
};

xQueueHandle xQueueSave;
xQueueHandle xQueueRecover;
xQueueHandle xQueueRecoverRequest;

static void vEDACTask(void *pvParameters);
static void vEDACSaveQueueHandler(void *pvParameters);
static void vEDACLoadQueueHandler(void *pvParameters);

static void vBasicTask1(void *pvParameters);
static void vBasicTask2(void *pvParameters);


// For flash test
//#define FLASHTEST
#if defined FLASHTEST
	#include "at45dbx.h"
	#include "conf_data_flash_example.h"
	#define	TARGET_SECTOR 0x00002
	uint8_t ram_buf[AT45DBX_SECTOR_SIZE];
	#define NUMBER_OF_SECTORS 4

	void flash_test_ko(void)
	{
		gpio_set_pin_high(DATA_FLASH_LED_EXAMPLE_1);
		while(1);
	}
#endif

xSemaphoreHandle xEDACDataSem;
struct StorageElement* storage_array_struct;
struct DataElement* data_array_struct;

int main(void)
{
	// initialize
	init();
	edac_init();
	at45dbx_init();
	// printf("Init done!");
	// fflush(stdout);
	
	
	// For testing of flash connection
	#if defined FLASHTEST
		uint16_t i;
		if(at45dbx_mem_check()==true)	{
			gpio_set_pin_low(DATA_FLASH_LED_EXAMPLE_0);
		} else
		{
			flash_test_ko();
		}
		// Prepare half a data flash sector to 0xAA
		for(i=0;i<AT45DBX_SECTOR_SIZE/2;i++) {
			ram_buf[i]=0xAA;
		}
		// And the remaining half to 0x55
		for(;i<AT45DBX_SECTOR_SIZE;i++) {
			ram_buf[i]=0x55;
		}

		at45dbx_write_sector_open(TARGET_SECTOR);
		at45dbx_write_sector_from_ram(ram_buf);
		at45dbx_write_close();

		// Read back this sector and compare to expected values
		at45dbx_read_sector_open(TARGET_SECTOR);
		at45dbx_read_sector_to_ram(ram_buf);
		at45dbx_read_close();
		for(i=0;i<AT45DBX_SECTOR_SIZE/2;i++) {
			if (ram_buf[i]!=0xAA) {
				flash_test_ko();
			}
		}
		for(;i<AT45DBX_SECTOR_SIZE;i++) {
			if (ram_buf[i]!=0x55) {
				flash_test_ko();
			}
		}

		// Write one data flash sector to 0x00, 0x01 ....
		for(i=0;i<AT45DBX_SECTOR_SIZE;i++) {
			ram_buf[i]=i;
		}
		at45dbx_write_sector_open(TARGET_SECTOR);
		at45dbx_write_sector_from_ram(ram_buf);
		at45dbx_write_close();

		// Read one data flash sector to ram
		at45dbx_read_sector_open(TARGET_SECTOR);
		at45dbx_read_sector_to_ram(ram_buf);
		at45dbx_read_close();
		for(i=0;i<AT45DBX_SECTOR_SIZE;i++) {
			if ( ram_buf[i]!=(i%0x100) ) {
				flash_test_ko();
			}
		}

		gpio_set_pin_low(DATA_FLASH_LED_EXAMPLE_1);
		while (1);
	#endif
	
	xEDACDataSem = xSemaphoreCreateMutex();
	if(xEDACDataSem == NULL){
		printf("Sem init ERROR\n");
		fflush(stdout);
	}
	
	// create queue for interprocess communication
	xQueueSave = xQueueCreate(queueLength, sizeof(struct AMessage));
	if(xQueueSave == 0){
		printf("QInit0E\n");
		fflush(stdout);
	}
	xQueueRecover = xQueueCreate(queueLength, sizeof(struct AMessage));
	if(xQueueRecover == 0){
		printf("QInit1E\n");
		fflush(stdout);
	}
	xQueueRecoverRequest = xQueueCreate(queueLength, sizeof(struct ARequest));
	if(xQueueRecover == 0){
		printf("QInit2E\n");
		fflush(stdout);
	}
	
	// Initializing BCH parameters
	edac_read_p();               // Read m
	edac_generate_gf();          // Construct the Galois Field GF(2**m)
	edac_gen_poly();             // Compute the generator polynomial of BCH code
	// Generate error field
	edac_gererate_errors();
	
	// Initializing data structures
	storage_array_struct = pvPortMalloc(edac_data_size*sizeof(struct StorageElement));
	for(i=0;i<edac_data_size;i++){
		storage_array_struct[i].data0 = 0;
		storage_array_struct[i].data1 = 0;
		storage_array_struct[i].data2 = 0;
	}
	data_array_struct = pvPortMalloc(edac_data_size*sizeof(struct DataElement));

	xTaskCreate( vEDACTask, (signed char * ) "BCH", configMINIMAL_STACK_SIZE+7500, NULL, tskIDLE_PRIORITY + 2, NULL);
	xTaskCreate( vEDACSaveQueueHandler, (signed char * ) "EdacQ", configMINIMAL_STACK_SIZE+200, NULL, tskIDLE_PRIORITY + 3, NULL);
	
	xTaskCreate( vBasicTask1, (signed char * ) "Basic1", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY + 1, NULL);
	xTaskCreate( vBasicTask2, (signed char * ) "Basic2", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY + 1, NULL);

	// Start the scheduler, anything after this will not run.
	vTaskStartScheduler();
}
static void vBasicTask1(void *pvParameters){
	const portTickType xDelay = 700 / portTICK_RATE_MS;
	uint8_t i=0;
	uint8_t send=1;
	
	// buffer for messages to queue
	struct AMessage *criticalVar = pvPortMalloc(sizeof(struct AMessage)); //vPortFree()
	criticalVar->qId = 1;
	criticalVar->qAddr = 0;
	criticalVar->qData[0] = 1;
	criticalVar->qData[1] = 1;
	criticalVar->qData[2] = 1;
	criticalVar->qData[3] = 1;
	criticalVar->qData[4] = 1;
	
	while(1){
		
		criticalVar->qAddr = i;
		// send message in queue
		if(send==1){
			if( xQueueSend( xQueueSave, ( void * ) &(criticalVar), ( portTickType ) 10 ) != pdPASS ){
				printf("Q1SendE\n");
			}else{
				//printf("Q1Send: %d \n",criticalVar->qAddr);
				i++;
			}
		}	
		if(i>=(edac_data_size/2)) send=0;
		
		vTaskDelay(xDelay);
	}	
}

static void vBasicTask2(void *pvParameters){
	const portTickType xDelay = 700 / portTICK_RATE_MS;
	uint8_t i=0;
	uint8_t send=1;
	
	// buffer for messages to queue
	struct AMessage *criticalVar = pvPortMalloc(sizeof(struct AMessage)); //vPortFree()
	criticalVar->ucMessageID = 0;
	criticalVar->qId = 2;
	criticalVar->qAddr = i+(edac_data_size/2);
	criticalVar->qData[0] = 2;
	criticalVar->qData[1] = 2;
	criticalVar->qData[2] = 2;
	criticalVar->qData[3] = 2;
	criticalVar->qData[4] = 2;
	
	while(1){
		
		criticalVar->qAddr = i+56;//(edac_data_size/2);
		// send message in queue
		if(send==1){
			
			if( xQueueSend( xQueueSave, ( void * ) &(criticalVar), ( portTickType ) 10 ) != pdPASS ){
				printf("Q2SendE\n");
			}else{
				//printf("Q2Send: %d \n", criticalVar->qAddr);
				i++;
			}
		}
		if(i>=(edac_data_size/2)) send=0;
		
		vTaskDelay(xDelay);
	}
}

int struct_write_address(struct DataElement* elem, int key){
	
	// Set adress to zero
	elem->data[edac_data_address_bit] = 0;
	elem->data[edac_data_address_bit-1] = 0;
	elem->data[edac_data_address_bit-2] = 0;
	elem->data[edac_data_address_bit-3] = 0;
	elem->data[edac_data_address_bit-4] = 0;
	elem->data[edac_data_address_bit-5] = 0;
	elem->data[edac_data_address_bit-6] = 0;
	elem->data[edac_data_address_bit-7] = 0;
	
	elem->data[edac_data_address_bit] = key & 0x1;
	elem->data[edac_data_address_bit-1] = (key & 0x2) >> 1;
	elem->data[edac_data_address_bit-2] = (key & 0x4) >> 2;
	elem->data[edac_data_address_bit-3] = (key & 0x8) >> 3;
	elem->data[edac_data_address_bit-4] = (key & 0x10) >> 4;
	elem->data[edac_data_address_bit-5] = (key & 0x20) >> 5;
	elem->data[edac_data_address_bit-6] = (key & 0x40) >> 6;
	elem->data[edac_data_address_bit-7] = (key & 0x80) >> 7;
	
	return 0;
}

// compare function for qsort()
int comp(const void * elem1, const void * elem2){
	struct StorageElement* a = (struct StorageElement*) elem1;
	struct StorageElement* b = (struct StorageElement*) elem2;
	
	if(a->data2 > b->data2)
		return 1;
	if(a->data2 < b->data2)
		return -1;
	
	return 0;
}	

void write_address(struct StorageElement* elem, int key){
	// 0b11100000 = 0xe0, set the last 5 bits to 0
	elem->data2 = elem->data2 & 0xe0;
	// 0b00011111 = 0x1f
	elem->data2 = elem->data2 | (key & 0x1f);
}

// pack data structures
void pack(struct StorageElement* array, struct EncodedElement* var){
	array->data0 = 0;
	array->data1 = 0;
	array->data2 = 0;
	
	uint8_t j;
	for(j=0;j<edac_encoded_data_element_size;j++){
		// printf("%d",var->encoded_data[j]);
		if(j<32){
			array->data0 += (var->encoded_data[j] << j);
		}
		else if(j<59){
			array->data1 += (var->encoded_data[j] << (j-32));
		}
		else{
			array->data2 += (var->encoded_data[j] << (66-j));
		}
	}
}
// unpack data structures
void unpack(struct EncodedElement* var, struct StorageElement* array){
	
	uint8_t j;
	for(j=0;j<edac_encoded_data_element_size;j++){
		//var->encoded_data[j] = data_input[j];
		
		if(j<32){
			var->encoded_data[j] = (array->data0 >> j)&0x1;
		}
		else if(j<59){
			var->encoded_data[j] = (array->data1 >> (j-32))&0x1;
		}
		else{
			var->encoded_data[j] = (array->data2 >> (66-j))&0x1;
		}
	}
}

static void vEDACSaveQueueHandler(void *pvParameters){

	uint8_t i;
	uint32_t flash_addr;
	
	struct AMessage *criticalVar = pvPortMalloc(sizeof(struct AMessage)); //vPortFree()
	criticalVar->qId = 3;
	criticalVar->qAddr = 3;
	criticalVar->qData[0] = 3;
	criticalVar->qData[1] = 3;
	criticalVar->qData[2] = 3;
	criticalVar->qData[3] = 3;
	criticalVar->qData[4] = 3;

	struct EncodedElement* encoded_struct;
	encoded_struct = pvPortMalloc(sizeof(struct EncodedElement));
	
	while(1){
		if(xQueueSave != 0){
			if( xQueueReceive( xQueueSave, &(criticalVar), ( portTickType ) 10 ) ){
				if(xSemaphoreTake(xEDACDataSem, (portTickType) 10) == pdTRUE){

					for(i=0;i<k;i++){
						if(i<8){
							data[i] = (criticalVar->qData[0] >> (7-i)) &0x1;
						}
						else if(i<16){
							data[i] = (criticalVar->qData[1] >> (15-i)) &0x1;
						}
						else if(i<24){
							data[i] = (criticalVar->qData[2] >> (23-i)) &0x1;
						}
						else if(i<32){
							data[i] = (criticalVar->qData[3] >> (31-i)) &0x1;
						}
						else if(i<40){
							data[i] = (criticalVar->qData[4] >> (39-i)) &0x1;
						}
						else if(i<45){
							data[i] = (criticalVar->qId >> (44-i)) &0x1;
						}
						else if(i<53){
							data[i] = (criticalVar->qAddr >> (52-i)) &0x1;
						}
						// Print for debug
						//printf("%d",data[i]);
					}
					//printf("\n");
					
					//flash_addr = ((criticalVar->qId << 8) &0x1f) + criticalVar->qAddr;
					flash_addr = criticalVar->qAddr;
				
					edac_encode_bch();           // encode data
				
					for (i = 0; i < length - k; i++){
						recd[i] = bb[i];
						encoded_struct->encoded_data[i] = bb[i];
					}
					for (i = 0; i < k; i++){
						recd[i + length - k] = data[i];
						encoded_struct->encoded_data[i + length - k] = data[i];
						data_array_struct[flash_addr].data[i] = data[i];
					}
				
					pack(&storage_array_struct[flash_addr], encoded_struct);
				
					xSemaphoreGive(xEDACDataSem);
				}
				else{
					printf("QRecvE\n");
				}
			}
		}
	}
}

static void vEDACLoadQueueHandler(void *pvParameters){
	uint8_t i;
	uint32_t flash_addr;

	struct AMessage *criticalVar = pvPortMalloc(sizeof(struct AMessage)); //vPortFree()
	criticalVar->qId = 3;
	criticalVar->qAddr = 3;
	criticalVar->qData[0] = 3;
	criticalVar->qData[1] = 3;
	criticalVar->qData[2] = 3;
	criticalVar->qData[3] = 3;
	criticalVar->qData[4] = 3;

	struct ARequest *recoverVar = pvPortMalloc(sizeof(struct ARequest));
	recoverVar->qId = 0;
	recoverVar->qAddr = 0;

	struct EncodedElement* encoded_struct;
	encoded_struct = pvPortMalloc(sizeof(struct EncodedElement));

	struct StorageElement** edac_storage_cache;
	edac_storage_cache = pvPortMalloc(edac_cache_size*sizeof(struct StorageElement*));
	for(i=0;i<edac_cache_size;i++){
		edac_storage_cache[i] = pvPortMalloc(2*sizeof(struct StorageElement));
	}
	struct StorageElement* bsearch_compare_struct;
	bsearch_compare_struct = pvPortMalloc(sizeof(struct StorageElement));
	struct StorageElement* bsearch_result_struct;
	i=5;
	write_address(bsearch_compare_struct,i);
	
	while(1){
		if(xQueueRecoverRequest != 0){
			if( xQueueReceive( xQueueRecoverRequest, &(recoverVar), ( portTickType ) 10 ) ){
				if(xSemaphoreTake(xEDACDataSem, (portTickType) 10) == pdTRUE){
					flash_addr = recoverVar->qAddr;
					unpack(encoded_struct, &storage_array_struct[flash_addr]);
					xSemaphoreGive(xEDACDataSem);

					// Format for sending
					for(i=0;i<5;i++){
						criticalVar->qData[i] = 0;
					}
					criticalVar->qId = 0;
					criticalVar->qAddr = 0;
					for (i = 0; i < k; i++){
						if(i<8){
							criticalVar->qData[0] += (encoded_struct->encoded_data[i + length - k] >> (7-i))&0x1;
							}else if(i<16){
							criticalVar->qData[1] += (encoded_struct->encoded_data[i + length - k] >> (15-i))&0x1;
							}else if(i<24){
							criticalVar->qData[2] += (encoded_struct->encoded_data[i + length - k] >> (23-i))&0x1;
							}else if(i<32){
							criticalVar->qData[3] += (encoded_struct->encoded_data[i + length - k] >> (31-i))&0x1;
							}else if(i<40){
							criticalVar->qData[4] += (encoded_struct->encoded_data[i + length - k] >> (39-i))&0x1;
							}else if(i<45){
							criticalVar->qId += (encoded_struct->encoded_data[i + length - k] >> (44-i))&0x1; // Would need to ensure use of 5 last bits if we did not previously set it to 0
							}else if(i<53){
							criticalVar->qAddr += (encoded_struct->encoded_data[i + length - k] >> (52-i))&0x1;
						}
						
					}
					
					// Put requested variable in queue
					if(xQueueRecover != 0){
						if( xQueueSend( xQueueRecover, ( void * ) &criticalVar, ( portTickType ) 10 ) != pdPASS ){
							printf("QRecoverE\n");
						}
					}
					else{
						printf("QRecoverE\n");
					}
				}
				else{
					printf("QRecoverRecvE\n");
				}
			}
		}
	}
}

static void vEDACTask(void *pvParameters)
{
	// Task delay time
	const portTickType xDelay = 5000 / portTICK_RATE_MS;
	
	// Counters
	int i;
	uint8_t j;
	
	// buffer for messages from queue
	struct AMessage *criticalVar = pvPortMalloc(sizeof(struct AMessage)); //vPortFree()
	criticalVar->qId = 3;
	criticalVar->qAddr = 3;
	criticalVar->qData[0] = 3;
	criticalVar->qData[1] = 3;
	criticalVar->qData[2] = 3;
	criticalVar->qData[3] = 3;
	criticalVar->qData[4] = 3;
	
	struct ARequest *recoverVar = pvPortMalloc(sizeof(struct ARequest));
	recoverVar->qId = 0;
	recoverVar->qAddr = 0;
	
	// variable to keep track of free space
	uint16_t edac_free_space = edac_data_size;
	uint8_t edac_free_space_map[edac_data_size]; // TODO: Ensure safe storage
	for(i=0;i<edac_data_size;i++){
		edac_free_space_map[i] = 1;
	}
		
	uint8_t edac_sort_freq = 8;
	uint8_t edac_counter = 0;
	uint8_t edac_data_sorted = 0;
	const uint8_t edac_correct_interval = 240;
	uint8_t edac_correct_freq = edac_correct_interval;
	
	// Flash data variables
	uint32_t flash_addr;
	uint16_t cache_counter = 0;
	
	// Allocating memory for encoded data
	struct EncodedElement* encoded_struct;
	encoded_struct = pvPortMalloc(sizeof(struct EncodedElement));
	
	struct StorageElement* storage_struct;
	storage_struct = pvPortMalloc(sizeof(struct StorageElement));
	
	// Initializing data and storage
	for(j=0;j<edac_data_size;j++){

		//fill data
		for (i = 0; i < k; i++){
			// Data array struct is for debug comparison of recovered data
			data_array_struct[j].data[i] = data_input[i];
					
			if(i == edac_data_element_size-1){
				struct_write_address(&data_array_struct[j], j);
			}
		}
		for (i = 0; i < k; i++){
			data[i] = data_array_struct[j].data[i];
			// debug print
			// printf("%d",data[i]);
		}
		//printf("\n");
				
		edac_encode_bch();           // encode data
				
		// recd[] are the coefficients of c(x) = x**(length-k)*data(x) + b(x)
				
		for (i = 0; i < length - k; i++){
			recd[i] = bb[i];
			encoded_struct->encoded_data[i] = bb[i];
		}
		for (i = 0; i < k; i++){
			recd[i + length - k] = data[i];
			encoded_struct->encoded_data[i + length - k] = data[i];
		}
				
				
		pack(&storage_array_struct[j], encoded_struct);
		printf("table %d content : ",j);
		printf("%u %u %u\n",(unsigned int) storage_array_struct[j].data0,
		(unsigned int) storage_array_struct[j].data1,
		(unsigned int) storage_array_struct[j].data2);
				
		// // Inject errors
		// numerr = errors[j][0];
		// for(i=0;i<numerr;i++){
			// 	errpos[i] = errors[j][i+1];
		// }
		// if (numerr){
			// 	for (i = 0; i < numerr; i++){
				// 		recd[errpos[i]] ^= 1;
				// 		encoded_array_struct[j].encoded_data[errpos[i]] ^= 1;
			// 	}
		// }
				
		for(i=0;i<67;i++){
			printf("%d",encoded_struct->encoded_data[i]);
		}
		printf("\n");
	}
	
	// Test flash
	
	while(1){
		
		if(edac_counter == edac_correct_freq && 1==2){
			// Write sectors to RAM for debug
			for(j=0;j<NUMBER_OF_SECTORS;j++){
				at45dbx_write_sector_open(TARGET_SECTOR+j);
				at45dbx_write_sector_from_ram(storage_array_struct);
				at45dbx_write_close();
			}
			// Write data to flash for compare
			for(j=20;j<NUMBER_OF_SECTORS+20;j++){
				at45dbx_write_sector_open(TARGET_SECTOR+j);
				at45dbx_write_sector_from_ram(data_array_struct);
				at45dbx_write_close();
			}	
			
			// Perform EDAC on memory
			printf("Performing EDAC on memory\n");
			
			for(j=0;j<NUMBER_OF_SECTORS;j++){
				// read sector from flash for decoding
				at45dbx_read_sector_open(TARGET_SECTOR+j);
				at45dbx_read_sector_to_ram(storage_array_struct);
				at45dbx_read_close();
			
				// Load one in, correct and write back
				for(j=0;j<edac_data_size;j++){

					if(xSemaphoreTake(xEDACDataSem, (portTickType) 10) == pdTRUE){
					
						unpack(encoded_struct, &storage_array_struct[j]);
				
						for(i=0;i<67;i++){
							recd[i] = encoded_struct->encoded_data[i];
						}
						edac_decode_bch();						
				
						// Write corrected data back to storage
						for(i=0;i<67;i++){
							encoded_struct->encoded_data[i] = recd[i];
						}
						pack(&storage_array_struct[j],encoded_struct);

						xSemaphoreGive(xEDACDataSem);
					}
				}
			
				at45dbx_write_sector_open(TARGET_SECTOR+j);
				at45dbx_write_sector_from_ram(storage_array_struct);
				at45dbx_write_close();
			}
			printf("EDAC Done\n");
			
			edac_correct_freq = edac_counter+edac_correct_interval;
		}
				
		gpio_toggle_pin(LED0_GPIO);

		
		// Done generating data and error arrays
		printf("\n");
		// Starting decoding
		if(xSemaphoreTake(xEDACDataSem, (portTickType) 10) == pdTRUE){
			printf("ETS\n");

			for(j=0;j<edac_data_size;j++){
				// Unpacking and writing
				unpack(encoded_struct, &storage_array_struct[j]);
				/* // Print for debug
				printf("unpack %d content : ",j);
				for(i=0;i<67;i++){
					printf("%d",encoded_struct->encoded_data[i]);
					if(i == edac_encoded_address_bit-5)
					printf(" ");
				}
				printf("\n");
				*/
			
				// Load data to recd for decode
				for(i=0;i<67;i++){
					recd[i] = encoded_struct->encoded_data[i];
				}
				edac_decode_bch();             // DECODE received codeword recv[] 
			
				// Liveliness led blink
				// gpio_toggle_pin(LED2_GPIO);
		
		
				// DECODING ERRORS? we compare only the data portion
				for (i = length - k; i < length; i++)
					if (data_array_struct[j].data[i - length + k] != recd[i])
						decerror++;
				if (decerror){
					printf("There were %d decoding errors in message positions:\t", decerror);
					fflush(stdout);
				}else{
					printf("Succesful decoding:\t");
					fflush(stdout);
				}
				//reset error counter
				decerror = 0;
			
				// Write corrected data back to storage
				for(i=0;i<67;i++){
					encoded_struct->encoded_data[i] = recd[i];
				}
				pack(&storage_array_struct[j],encoded_struct);
			
				// Print written data for debug
				for(i=0;i<67;i++){
					printf("%d",encoded_struct->encoded_data[i]);
					if(i == edac_encoded_address_bit-8)
						printf(" ");
				}
				printf("  %d \n",	(encoded_struct->encoded_data[edac_encoded_address_bit-7] << 7)+
									(encoded_struct->encoded_data[edac_encoded_address_bit-6] << 6)+
									(encoded_struct->encoded_data[edac_encoded_address_bit-5] << 5)+
									(encoded_struct->encoded_data[edac_encoded_address_bit-4] << 4)+
									(encoded_struct->encoded_data[edac_encoded_address_bit-3] << 3)+
									(encoded_struct->encoded_data[edac_encoded_address_bit-2] << 2)+
									(encoded_struct->encoded_data[edac_encoded_address_bit-1] << 1)+
									(encoded_struct->encoded_data[edac_encoded_address_bit]));
			}
			printf("EGS\n");
			xSemaphoreGive(xEDACDataSem);
		}		
		// Counter to determine frequency of specific task operations
		edac_counter++;
		
		/* // sorting of cached variables
		if(edac_counter % edac_sort_freq == 0 && edac_data_sorted == 0){
			// Sort memory area
			
			//qsort(&encoded_array_struct[0], edac_data_size, sizeof(struct EncodedElement), struct_comp);
			qsort(&storage_array_struct[0], edac_data_size, sizeof(struct StorageElement), comp);
			edac_data_sorted = 1;
			
			// Seach memory area
			//bsearch_result_struct = (struct EncodedElement*) bsearch(bsearch_compare_struct,&encoded_array_struct[0], edac_data_size, sizeof(struct EncodedElement), struct_comp);
			bsearch_result_struct = (struct StorageElement*) bsearch(bsearch_compare_struct,&storage_array_struct[0], edac_data_size, sizeof(struct StorageElement), comp);
			if(bsearch_result_struct != NULL){
				printf("Search found item:\n");
				printf("%d",bsearch_result_struct->data2 & 0x1f);
				//for(j=0;j<edac_encoded_data_element_size;j++){
				//	printf("%d",bsearch_result_struct->encoded_data[j]);
				//}
				printf("\n");
			}
			else{
				printf("Search did not find item:\n");
			}
			
		}
		*/
		
		printf("Iteration complete\n\n");
		gpio_toggle_pin(LED1_GPIO);
		
		vTaskDelay(xDelay);
	}
}