/*This file has been prepared for Doxygen automatic documentation generation.*/
/*! \file *********************************************************************
 *
 * \brief FreeRTOS Real Time Kernel example.
 *
 * Creates all the demo application tasks, then starts the scheduler.  The WEB
 * documentation provides more details of the demo application tasks.
 *
 * Main. c also creates a task called "Check".  This only executes every three
 * seconds but has the highest priority so is guaranteed to get processor time.
 * Its main function is to check that all the other tasks are still operational.
 * Each task that does not flash an LED maintains a unique count that is
 * incremented each time the task successfully completes its function.  Should
 * any error occur within such a task the count is permanently halted.  The
 * check task inspects the count of each task to ensure it has changed since
 * the last time the check task executed.  If all the count variables have
 * changed all the tasks are still executing error free, and the check task
 * toggles an LED.  Should any task contain an error at any time the LED toggle
 * will stop.
 *
 * The LED flash and communications test tasks do not maintain a count.
 *
 * - Compiler:           IAR EWAVR32 and GNU GCC for AVR32
 * - Supported devices:  All AVR32 devices with GPIO.
 * - AppNote:
 *
 * \author               Atmel Corporation: http://www.atmel.com \n
 *                       Support and FAQ: http://support.atmel.no/
 *
 *****************************************************************************/

/*
    FreeRTOS V7.0.0 - Copyright (C) 2011 Real Time Engineers Ltd.
	

    ***************************************************************************
     *                                                                       *
     *    FreeRTOS tutorial books are available in pdf and paperback.        *
     *    Complete, revised, and edited pdf reference manuals are also       *
     *    available.                                                         *
     *                                                                       *
     *    Purchasing FreeRTOS documentation will not only help you, by       *
     *    ensuring you get running as quickly as possible and with an        *
     *    in-depth knowledge of how to use FreeRTOS, it will also help       *
     *    the FreeRTOS project to continue with its mission of providing     *
     *    professional grade, cross platform, de facto standard solutions    *
     *    for microcontrollers - completely free of charge!                  *
     *                                                                       *
     *    >>> See http://www.FreeRTOS.org/Documentation for details. <<<     *
     *                                                                       *
     *    Thank you for using FreeRTOS, and thank you for your support!      *
     *                                                                       *
    ***************************************************************************


    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.
    >>>NOTE<<< The modification to the GPL is included to allow you to
    distribute a combined work that includes FreeRTOS without being obliged to
    provide the source code for proprietary components outside of the FreeRTOS
    kernel.  FreeRTOS is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details. You should have received a copy of the GNU General Public
    License and the FreeRTOS license exception along with FreeRTOS; if not it
    can be viewed here: http://www.freertos.org/a00114.html and also obtained
    by writing to Richard Barry, contact details for whom are available on the
    FreeRTOS WEB site.

    1 tab == 4 spaces!

    http://www.FreeRTOS.org - Documentation, latest information, license and
    contact details.

    http://www.SafeRTOS.com - A version that is certified for use in safety
    critical systems.

    http://www.OpenRTOS.com - Commercial support, development, porting,
    licensing and training services.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <intc.h>
#include <interrupt.h>
#include <sleepmgr.h>
#include <uc3/sleepmgr.h>
#include <gpio.h>
#include <flashc.h>
#include <power_clocks_lib.h>
#include <sleep.h>
#include <sysclk.h>
#include <tc.h>
#include <usart.h>
#include <udi_cdc.h>
#include <usb_protocol_cdc.h>
#include <udc.h>
#include <udd.h>
#include <compiler.h>
#include <status_codes.h>
#include <board.h>
#include <stdio_usb.h>


// defines for USB UART
#define CONFIG_USART_IF (AVR32_USART2)

// include FreeRTOS header files
#include "FreeRTOS.h"
#include "task.h"

void init(void)
{
	// board init
	board_init();
	
#if UC3L
	// clock frequencies
	#define EXAMPLE_TARGET_DFLL_FREQ_HZ     96000000  // DFLL target frequency, in Hz
	#define EXAMPLE_TARGET_MCUCLK_FREQ_HZ   12000000  // MCU clock target frequency, in Hz
	#define EXAMPLE_TARGET_PBACLK_FREQ_HZ   12000000  // PBA clock target frequency, in Hz

	// parameters to pcl_configure_clocks().
	static scif_gclk_opt_t gc_dfllif_ref_opt = { SCIF_GCCTRL_SLOWCLOCK, 0, false};
	static pcl_freq_param_t pcl_dfll_freq_param = {
		.main_clk_src = PCL_MC_DFLL0,
		.cpu_f        = EXAMPLE_TARGET_MCUCLK_FREQ_HZ,
		.pba_f        = EXAMPLE_TARGET_PBACLK_FREQ_HZ,
		.pbb_f        = EXAMPLE_TARGET_PBACLK_FREQ_HZ,
		.dfll_f       = EXAMPLE_TARGET_DFLL_FREQ_HZ,
		.pextra_params = &gc_dfllif_ref_opt
	};
	pcl_configure_clocks(&pcl_dfll_freq_param);
#else
	pcl_switch_to_osc(PCL_OSC0, FOSC0, OSC0_STARTUP);
#endif	
	
	// stdio init
	stdio_usb_init(&CONFIG_USART_IF);

	// Specify that stdout and stdin should not be buffered.

#if defined(__GNUC__) && defined(__AVR32__)
	setbuf(stdout, NULL);
	setbuf(stdin,  NULL);
#endif
}

#include <semphr.h>
//#define MASTER
#define TIMEOUT 20 // 20*100ms is 2sec

#if defined (MASTER)
static void vMasterTask(void *pvParameters);

#define RESET_S			(AVR32_PIN_PA31) //GPIO0
#define HEARTBEAT_M		(AVR32_PIN_PA29) //GPIO2
#define HEARTBEAT_S		(AVR32_PIN_PA27) //GPIO4
#else

static void vSlaveTask(void *pvParameters);
static void vAssumeMasterTask(void *pvParameters);

#define RESET_M			(AVR32_PIN_PA31) //GPIO0
#define HEARTBEAT_M		(AVR32_PIN_PA29) //GPIO3
#define HEARTBEAT_S		(AVR32_PIN_PA27) //GPIO5
#endif

xSemaphoreHandle xMasterFlagSem;
uint8_t masterFlag;


void private_init(void){
	masterFlag = 0;
	
	#if defined (MASTER)
	gpio_configure_pin (HEARTBEAT_S,GPIO_DIR_INPUT);
	gpio_configure_pin (HEARTBEAT_M,GPIO_DIR_OUTPUT | GPIO_INIT_HIGH);
	gpio_configure_pin (RESET_S,	GPIO_DIR_OUTPUT | GPIO_INIT_HIGH);
	#else
	gpio_configure_pin (HEARTBEAT_S,GPIO_DIR_OUTPUT | GPIO_INIT_HIGH);
	gpio_configure_pin (HEARTBEAT_M,GPIO_DIR_INPUT);
	gpio_configure_pin (RESET_M,	GPIO_DIR_OUTPUT | GPIO_INIT_HIGH);
	
	xMasterFlagSem = xSemaphoreCreateMutex();
	if(xMasterFlagSem == NULL){
		printf("Sem init ERROR\n");
		fflush(stdout);
	}
	#endif
}

int main(void)
{
	// initialize
	init();
	private_init();
	
	// start master task
	#if defined (MASTER)
	xTaskCreate( vMasterTask, (signed char * ) "MASTER", configMINIMAL_STACK_SIZE+1000, NULL, tskIDLE_PRIORITY + 10, NULL);
	// start slave task
	#else
	xTaskCreate( vSlaveTask, (signed char * ) "SLAVE", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY + 10, NULL);
	xTaskCreate( vAssumeMasterTask, (signed char * ) "AM", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY + 1, NULL);
	#endif
	
	// Start the scheduler
	gpio_toggle_pin(LED3_GPIO);
	vTaskStartScheduler();
}

#if defined (MASTER)
static void vMasterTask(void *pvParameters)
{
	const portTickType xDelay = 100 / portTICK_RATE_MS;
	
	uint8_t heartbeat = 0;
	uint8_t heartbeat_prev = 0;
	
	uint8_t counter = 0;
	
	while(1)
	{
		gpio_toggle_pin(LED0_GPIO);			// Alive blink
		gpio_toggle_pin(HEARTBEAT_M);
		heartbeat = gpio_get_pin_value(HEARTBEAT_S);	// Read slave heartbeat
		
		if(heartbeat_prev ^ heartbeat){
			counter = 0;
		}
		
		if (counter > (TIMEOUT+10)){
			gpio_set_pin_high(RESET_S);
			counter = 0;
		}
		else if(counter >= TIMEOUT){
			gpio_set_pin_low(RESET_S);		//reset slave
			//gpio_toggle_pin(LED2_GPIO);
		}
		counter++;
		//printf("tick, counter: %d\n",counter);
		//fflush(stdout);
		//gpio_toggle_pin(LED1_GPIO);
		heartbeat_prev = heartbeat;
		counter++;
		vTaskDelay(xDelay);
	}
}
#else
static void vSlaveTask(void *pvParameters)
{
	const portTickType xDelay = 100 / portTICK_RATE_MS;
	
	uint8_t heartbeat = 0;
	uint8_t heartbeat_prev = 0;
	
	uint8_t counter = 0;
	
	while(1)
	{
		gpio_toggle_pin(LED0_GPIO);			// Alive blink
		gpio_toggle_pin(HEARTBEAT_S);
		heartbeat = gpio_get_pin_value(HEARTBEAT_M);	// Read slave heartbeat
		
		if(heartbeat_prev ^ heartbeat){
			counter = 0;
		}
		
		if (counter > (TIMEOUT+10)){
			gpio_set_pin_high(RESET_M);
			counter = 0;
		}
		else if(counter >= TIMEOUT){
			gpio_set_pin_low(RESET_M);		//reset slave
			if(xSemaphoreTake(xMasterFlagSem,(portTickType) 10) == pdTRUE){
				masterFlag = 1;
				xSemaphoreGive(xMasterFlagSem);
			}
		}
		if(heartbeat_prev ^ heartbeat){
			if(xSemaphoreTake(xMasterFlagSem,(portTickType) 10) == pdTRUE){
				masterFlag = 0;
				xSemaphoreGive(xMasterFlagSem);
			}
		}
		//printf("tick, counter: %d\n",counter);
		//fflush(stdout);
		//gpio_toggle_pin(LED1_GPIO);
		heartbeat_prev = heartbeat;
		counter++;
		vTaskDelay(xDelay);
	}
}

static void vAssumeMasterTask(void *pvParameters)
{
	const portTickType xDelay = 1000 / portTICK_RATE_MS;
	
	while(1){
		if(xSemaphoreTake(xMasterFlagSem,(portTickType) 10) == pdTRUE){
			if(masterFlag == 0){
				vTaskDelay(xDelay);
			}else{
				gpio_toggle_pin(LED1_GPIO);
			}
			xSemaphoreGive(xMasterFlagSem);
			
			//collect data and do master duties
			gpio_toggle_pin(LED2_GPIO);
			vTaskDelay(xDelay);
		}
	}
}

#endif